<?php

namespace HtmlDriven\CorsProxy;

use DateTime;
use Dibi\Connection as DibiConnection;
use Guzzle\Http\Client;
use Guzzle\Http\ClientInterface;
use HtmlDriven\CorsProxy\Config;
use HtmlDriven\CorsProxy\Exception\FileNotFoundException;
use HtmlDriven\CorsProxy\Exception\MalformedConfigFileException;
use PDO;
use RuntimeException;

/**
 * Main application entry point.
 *
 * @author RebendaJiri <jiri.rebenda@htmldriven.com>
 */
class Application
{
    /** @var string */
    const VERSION = '1.1.0';

    /** @var string */
    const DEFAULT_PROTOCOL = 'http';

    /** @var string */
    const DEFAULT_DOMAIN = 'cors-proxy.htmldriven.com';

    /** @var string */
    const ACCESS_TOKEN_PARAMETER_NAME = 't';

    /** @var array */
    public $configDefaults = [];

    /** @var Config */
    private $config;

    /** @var string[] */
    private $allowedActions = [
        'frontend',
        'sitemap',
    ];

    /**
     * @param Config|NULL If NULL, default config is created
     */
    public function __construct(Config $config = null)
    {
        $this->configDefaults = $this->createConfigDefaults();
        $this->config = ($config === null ? $this->createConfig() : $config);
    }

    /**
     * Handles the request.
     *
     * @return void
     * @throws MalformedConfigFileException If configuration file parsing fails.
     * @throws FileNotFoundException If template file does not exist.
     * @throws RuntimeException If HTTP method cannot be detected.
     */
    public function run()
    {
        $config = $this->config;

        // handle CORS proxy request if URL is set
        if (isset($_GET[$config->getUrlParameterName()])) {
            $isRequestValid = true;

            // IP blacklist
            if ($config->getIPBlacklistFile() !== null) {
                if (isset($_SERVER['REMOTE_ADDR'])) {
                    $ipBlacklist = IPBlacklist::loadFromFile($config->getIPBlacklistFile());
                    if ($ipBlacklist->isIPBlacklisted($_SERVER['REMOTE_ADDR'])) {
                        $isRequestValid = false;

                        http_response_code(403);
                        header('Content-Type: application/json');
                        $json = [
                            'success' => false,
                            'error' => 'You are not allowed to use this service.',
                            'body' => null,
                        ];
                        echo json_encode($json);
                    }
                }
            }

            // access tokens whitelisting
            if ($isRequestValid && $config->getAccessTokensFile() !== null) {
                $accessTokens = AccessTokens::loadFromFile($config->getAccessTokensFile());

                if (!isset($_GET[static::ACCESS_TOKEN_PARAMETER_NAME])
                    || !$accessTokens->isAccessTokenValid($_GET[static::ACCESS_TOKEN_PARAMETER_NAME])) {
                    $isRequestValid = false;

                    http_response_code(401);
                    header('Content-Type: application/json');
                    $json = [
                        'success' => false,
                        'error' => 'You are not allowed to use this service without valid access token.',
                        'body' => null,
                    ];
                    echo json_encode($json);
                }
            }

            if ($isRequestValid) {
                $url = (string) $_GET[$config->getUrlParameterName()];

                $client = $this->createClient($config);

                $dibiConnection = $this->createDibiConnection($config);

                $requestHandler = new RequestHandler($client, $dibiConnection);

                $dibiConnection->disconnect();

                $method = $this->detectHttpMethod();

                $requestHandler->handleRequest($method, $url);
            }
        } else {
            // show frontend
            $this->showFrontEnd($config);
        }
    }

    /**
     * @return Config
     * @throws MalformedConfigFileException If configuration file parsing fails.
     * @throws FileNotFoundException If template file does not exist.
     */
    private function createConfig()
    {
        // config file is optional
        $iniFile = __DIR__ . '/../app/config/config.ini';

        if (is_file($iniFile)) {
            if (false === ($iniConfig = parse_ini_file($iniFile, true))) {
                throw new MalformedConfigFileException(sprintf("Unable to parse configuration file '%s'.", $iniFile));
            }
            $config = array_replace_recursive($this->configDefaults, $iniConfig);
        } else {
            $config = $this->configDefaults;
        }

        date_default_timezone_set($config['timezone']);

        // absolutize filepaths
        $config['templateFile'] = Helpers::absolutizeFilepath(__DIR__, $config['templateFile']);
        $config['sitemapTemplateFile'] = Helpers::absolutizeFilepath(__DIR__, $config['sitemapTemplateFile']);
        $config['errorTemplateFile'] = Helpers::absolutizeFilepath(__DIR__, $config['errorTemplateFile']);

        if ($config['ipBlacklistFile'] === 'false' || trim($config['ipBlacklistFile']) === '') {
            $config['ipBlacklistFile'] = null;
        }

        if ($config['accessTokensFile'] === 'false' || trim($config['accessTokensFile']) === '') {
            $config['accessTokensFile'] = null;
        }

        return new Config(
            $config['urlParameterName'],
            $config['userAgent'],
            $config['templateFile'],
            $config['sitemapPath'],
            $config['sitemapTemplateFile'],
            $config['errorTemplateFile'],
            $config['ipBlacklistFile'],
            $config['accessTokensFile'],
            $config['database']
        );
    }

    /**
     * @param Config
     * @return ClientInterface
     */
    private function createClient(Config $config)
    {
        $client = new Client();
        $client->setUserAgent($config->getUserAgent());
        return $client;
    }

    /**
     * @param Config
     * @return DibiConnection
     */
    private function createDibiConnection(Config $config)
    {
        $dibiConnection = new DibiConnection(
            $config->getDatabaseConfig()
        );
        return $dibiConnection;
    }

    /**
     * Detects HTTP method from global server params.
     *
     * @return string
     * @throws RuntimeException If HTTP method cannot be detected.
     */
    private function detectHttpMethod()
    {
        if (!isset($_SERVER['REQUEST_METHOD'])) {
            throw new RuntimeException('Unable to detect HTTP method.');
        }
        return strtoupper($_SERVER['REQUEST_METHOD']);
    }

    /**
     * @param Config
     * @return void
     */
    private function showFrontEnd(Config $config)
    {
        $isSecured = $this->isSecured();
        $domain = isset($_SERVER['HTTP_HOST']) ? $_SERVER['HTTP_HOST'] : static::DEFAULT_DOMAIN;
        $protocol = ($isSecured ? 'https' : 'http');
        $domainUrl = $protocol . '://' . $domain;
        $basePath = '';

        $lastUpdate = $this->getLastUpdate($config);

        $action = null;

        $error = null;

        $requestUri = isset($_SERVER['REQUEST_URI']) ? rtrim($_SERVER['REQUEST_URI'], '/') : null;
        if (isset($requestUri)) {
            if ($requestUri === '/index.php') {
                header('HTTP/1.1 301 Moved Permanently');
                header('Location: ' . $domainUrl . $basePath);
                return;
            } elseif ($requestUri === '') {
                $action = 'frontend';
            } elseif ($requestUri === $config->getSitemapPath()) {
                $action = 'sitemap';
            }
        } else {
            $error = 500;
            header('HTTP/1.1 500 Internal Server Error');
        }

        if (!in_array($action, $this->allowedActions, true)) {
            $action = 'error';
            $error = 404;
            header('HTTP/1.1 404 Not Found');
        }

        switch ($action) {
            case 'sitemap':
                require $config->getSitemapTemplateFile();
                break;
            case 'frontend':
                require $config->getTemplateFile();
                break;
            case 'error':
                require $config->getErrorTemplateFile();
                break;
        }
    }

    /**
     * Checks whether HTTPs is used.
     *
     * @return bool
     */
    private function isSecured()
    {
        return (!empty($_SERVER['HTTPS']) && $_SERVER['HTTPS'] !== 'off')
            || (isset($_SERVER['SERVER_PORT']) && $_SERVER['SERVER_PORT'] == 443);
    }

    /**
     * @param Config
     * @return DateTime
     */
    private function getLastUpdate(Config $config)
    {
        return new DateTime(date('Y-m-d H:i:s', filemtime($config->getTemplateFile())));
    }

    /**
     * @return array
     */
    private function createConfigDefaults()
    {
        return [
            'urlParameterName' => 'url',
            'userAgent' => 'htmldriven/cors-proxy ' . static::VERSION,
            'templateFile' => __DIR__ . '/../app/templates/default/frontend.phtml',
            'sitemapPath' => '/sitemap.xml',
            'sitemapTemplateFile' => __DIR__ . '/../app/templates/default/sitemap.pxml',
            'errorTemplateFile' => __DIR__ . '/../app/templates/default/error.phtml',
            'ipBlacklistFile' => false,
            'accessTokensFile' => false,
            'timezone' => 'UTC',
            'database' => [
                'driver' => 'pdo',
                'dsn' => 'mysql:dbname=cors_proxy;host=' . $this->detectMySqlHost() . ';charset=utf8mb4',
                'username' => 'cors_proxy',
                'password' => '',
                'lazy' => true,
                'timezone' => '+00:00',
                'options' => [
                    PDO::MYSQL_ATTR_INIT_COMMAND => 'SET time_zone = \'+00:00\'',
                    PDO::ATTR_ERRMODE => PDO::ERRMODE_EXCEPTION,
                ],
            ],
        ];
    }

    /**
     * Detects MySQL host.
     *
     * @return string
     */
    private function detectMySqlHost()
    {
        $host = getenv('_MYSQL_HOST');
        if ($host === false) {
            $host = '127.0.0.1';
        }
        return $host;
    }
}
